@@ -23,13 +23,16 @@ import datetime
 import gi
 gi.require_version('Gtk', '3.0')
 
-from gi.repository import GObject, GLib
+from gi.repository import GObject, GLib, Gio
 from gi.repository import Gtk, Gdk, GdkPixbuf
 
 import dbus
 import dbus.service
 from dbus.mainloop.glib import DBusGMainLoop
-import mlt
+try:
+    import mlt7 as mlt
+except:
+    import mlt
 import hashlib
 import locale
 import os
@@ -52,10 +55,8 @@ import editorstate
 import editorpersistance
 import gui
 import guiutils
-import mltenv
+import mltinit
 import mltprofiles
-import mlttransitions
-import mltfilters
 import processutils
 import persistance
 import respaths
@@ -86,6 +87,7 @@ UNQUEUED = 3
 ABORTED = 4
 
 render_queue = []
+_batch_render_app = None
 batch_window = None
 render_thread = None
 queue_runner_thread = None
@@ -96,6 +98,7 @@ _dbus_service = None
 
 render_item_menu = Gtk.Menu()
 
+_single_render_app = None
 single_render_window = None
 single_render_launch_thread = None
 single_render_thread = None
@@ -124,6 +127,8 @@ class QueueRunnerThread(threading.Thread):
 
             project = persistance.load_project(project_file_path, False)
 
+            project.c_seq.fix_v1_for_render()
+
             maybe_create_render_folder(render_item.render_path)
         
             producer = project.c_seq.tractor
@@ -144,10 +149,7 @@ class QueueRunnerThread(threading.Thread):
             # Set render start time and item state
             render_item.render_started()
 
-            Gdk.threads_enter()
-            batch_window.update_queue_view()
-            batch_window.current_render.set_text("  " + render_item.get_display_name())
-            Gdk.threads_leave()
+            GLib.idle_add(self._render_start_update, render_item)
 
             # Make sure that render thread is actually running before
             # testing render_thread.running value later
@@ -163,17 +165,13 @@ class QueueRunnerThread(threading.Thread):
                 render_fraction = render_thread.get_render_fraction()
                 now = time.time()
                 current_render_time = now - render_item.start_time
-                
-                Gdk.threads_enter()
-                batch_window.update_render_progress(render_fraction, items, render_item.get_display_name(), current_render_time)
-                Gdk.threads_leave()
+    
+                GLib.idle_add(self._render_progress_update, render_fraction, items, render_item.get_display_name(), current_render_time)
                 
                 if render_thread.running == False: # Rendering has reached end
                     self.thread_running = False
                     
-                    Gdk.threads_enter()
-                    batch_window.render_progress_bar.set_fraction(1.0)
-                    Gdk.threads_leave()
+                    GLib.idle_add(self._progress_bar_update, 1.0)
                                     
                     render_item.render_completed()
                 else:
@@ -181,9 +179,7 @@ class QueueRunnerThread(threading.Thread):
                     
             if not self.aborted:
                 items = items + 1
-                Gdk.threads_enter()
-                batch_window.update_render_progress(0, items, render_item.get_display_name(), 0)
-                Gdk.threads_leave()
+                GLib.idle_add(self._render_progress_update, 0, items, render_item.get_display_name(), 0)
             else:
                 if render_item != None:
                     render_item.render_aborted()
@@ -191,11 +187,24 @@ class QueueRunnerThread(threading.Thread):
             render_thread.shutdown()
         
         # Update view for render end
-        Gdk.threads_enter()
+        GLib.idle_add(self._queue_done_update)
+
+    def _render_start_update(self, render_item):
+        batch_window.update_queue_view()
+        batch_window.current_render.set_text("  " + render_item.get_display_name())
+        batch_window.current_file.set_text("  " +  os.path.basename(render_item.render_path))
+
+    def _render_progress_update(self, render_fraction, items, display_time, current_render_time):
+        batch_window.update_render_progress(render_fraction, items, display_time, current_render_time)
+
+    def _progress_bar_update(self, fraction):
+        batch_window.render_progress_bar.set_fraction(fraction)
+
+    def _queue_done_update(self):
+        # Update view for render end
         batch_window.reload_queue() # item may havee added to queue while rendering
         batch_window.render_queue_stopped()
-        Gdk.threads_leave()
-                    
+        
     def abort(self):
         render_thread.shutdown()
         # It may be that 'aborted' and 'running' could combined into single flag, but whatevaar
@@ -259,8 +268,6 @@ def add_render_item(flowblade_project, render_path, args_vals_list, mark_in, mar
     else:
         launch_batch_rendering()
 
-    #print "Render queue item for rendering file into " + render_path + " with identifier " + identifier + " added."
-
 # ------------------------------------------------------- file utils
 def init_dirs_if_needed():
     user_dir = userfolders.get_cache_dir()
@@ -341,63 +348,45 @@ def main(root_path, force_launch=False):
 
     # Load editor prefs and list of recent projects
     editorpersistance.load()
+
+    # Create app.
+    global _batch_render_app
+    _batch_render_app = BatchRenderApp()
+    _batch_render_app.run(None)
+
     
-    # Init translations module with translations data
-    translations.init_languages()
-    translations.load_filters_translations()
-    mlttransitions.init_module()
+class BatchRenderApp(Gtk.Application):
+    def __init__(self, *args, **kwargs):
+        Gtk.Application.__init__(self, application_id=None,
+                                 flags=Gio.ApplicationFlags.FLAGS_NONE)
+        self.connect("activate", self.on_activate)
 
-    # Init gtk threads
-    Gdk.threads_init()
-    Gdk.threads_enter()
-       
-    editorpersistance.load()
-    if editorpersistance.prefs.theme != appconsts.LIGHT_THEME:
-        Gtk.Settings.get_default().set_property("gtk-application-prefer-dark-theme", True)
-        if editorpersistance.prefs.theme == appconsts.FLOWBLADE_THEME:
-            gui.apply_gtk_css()
-            
-    repo = mlt.Factory().init()
-    processutils.prepare_mlt_repo(repo)
-    
-    # Set numeric locale to use "." as radix, MLT initilizes this to OS locale and this causes bugs 
-    locale.setlocale(locale.LC_NUMERIC, 'C')
+    def on_activate(self, data=None):
+        gui.apply_theme(editorpersistance.prefs.theme)
 
-    # Check for codecs and formats on the system
-    mltenv.check_available_features(repo)
-    renderconsumer.load_render_profiles()
+        mltinit.init_with_translations()
+        
+        global render_queue
+        render_queue = RenderQueue()
+        render_queue.load_render_items()
 
-    # Load filter and compositor descriptions from xml files.
-    mltfilters.load_filters_xml(mltenv.services)
-    mlttransitions.load_compositors_xml(mltenv.transitions)
+        global batch_window
+        batch_window = BatchRenderWindow()
 
-    # Create list of available mlt profiles
-    mltprofiles.load_profile_list()
+        if render_queue.error_status != None:
+            primary_txt = _("Error loading render queue items!")
+            secondary_txt = _("Message:\n") + render_queue.get_error_status_message()
+            dialogutils.warning_message(primary_txt, secondary_txt, batch_window.window)
 
-    global render_queue
-    render_queue = RenderQueue()
-    render_queue.load_render_items()
-
-    global batch_window
-    batch_window = BatchRenderWindow()
-
-    if render_queue.error_status != None:
-        primary_txt = _("Error loading render queue items!")
-        secondary_txt = _("Message:\n") + render_queue.get_error_status_message()
-        dialogutils.warning_message(primary_txt, secondary_txt, batch_window.window)
-
-    DBusGMainLoop(set_as_default=True)
-    global _dbus_service
-    _dbus_service = BatchRenderDBUSService()
-
-    Gtk.main()
-    Gdk.threads_leave()
+        DBusGMainLoop(set_as_default=True)
+        global _dbus_service
+        _dbus_service = BatchRenderDBUSService()
 
+        self.add_window(batch_window.window)
+        
 def _show_single_instance_info():
     global timeout_id
-    timeout_id = GObject.timeout_add(200, _display_single_instance_window)
-    # Launch gtk+ main loop
-    Gtk.main()
+    timeout_id = GLib.timeout_add(200, _display_single_instance_window)
     
 def _display_single_instance_window():
     GObject.source_remove(timeout_id)
@@ -421,7 +410,6 @@ def _display_single_instance_window():
 
 def _early_exit(dialog, response):
     dialog.destroy()
-    Gtk.main_quit() 
     
 def shutdown():
     if queue_runner_thread != None:
@@ -430,12 +418,14 @@ def shutdown():
         dialogutils.info_message(primary_txt, secondary_txt, batch_window.window)
         return True # Tell callsite (inside GTK toolkit) that event is handled, otherwise it'll destroy window anyway.
 
+    # GTK4 
     while(Gtk.events_pending()):
         Gtk.main_iteration()
 
     if _dbus_service != None:
         _dbus_service.remove_from_dbus()
-    Gtk.main_quit()
+
+    _batch_render_app.quit()
 
 
 class RenderQueue:
@@ -656,15 +646,19 @@ class BatchRenderWindow:
         self.est_time_left = Gtk.Label()
         self.current_render = Gtk.Label()
         self.current_render_time = Gtk.Label()
+        self.current_file = Gtk.Label()
         est_r = guiutils.get_right_justified_box([guiutils.bold_label(_("Estimated Left:"))])
         current_r = guiutils.get_right_justified_box([guiutils.bold_label(_("Current Render:"))])
         current_r_t = guiutils.get_right_justified_box([guiutils.bold_label(_("Elapsed:"))])
+        current_file = guiutils.get_right_justified_box([guiutils.bold_label(_("File:"))])
         est_r.set_size_request(250, 20)
         current_r.set_size_request(250, 20)
         current_r_t.set_size_request(250, 20)
+        current_file.set_size_request(250, 20)
         
         info_vbox = Gtk.VBox(False, 0)
         info_vbox.pack_start(guiutils.get_left_justified_box([current_r, self.current_render]), False, False, 0)
+        info_vbox.pack_start(guiutils.get_left_justified_box([current_file, self.current_file]), False, False, 0)
         info_vbox.pack_start(guiutils.get_left_justified_box([current_r_t, self.current_render_time]), False, False, 0)
         info_vbox.pack_start(guiutils.get_left_justified_box([est_r, self.est_time_left]), False, False, 0)
         
@@ -681,16 +675,16 @@ class BatchRenderWindow:
         self.render_progress_bar = Gtk.ProgressBar()
         self.render_progress_bar.set_text(self.not_rendering_txt)
 
-        self.remove_selected = Gtk.Button(_("Delete Selected"))
+        self.remove_selected = Gtk.Button(label=_("Delete Selected"))
         self.remove_selected.connect("clicked", 
                                      lambda w, e: self.remove_selected_clicked(), 
                                      None)
-        self.remove_finished = Gtk.Button(_("Delete Finished"))
+        self.remove_finished = Gtk.Button(label=_("Delete Finished"))
         self.remove_finished.connect("clicked", 
                                      lambda w, e: self.remove_finished_clicked(), 
                                      None)
 
-        self.reload_button = Gtk.Button(_("Reload Queue"))
+        self.reload_button = Gtk.Button(label=_("Reload Queue"))
         self.reload_button.connect("clicked", 
                                      lambda w, e: self.reload_queue(), 
                                      None)
@@ -701,7 +695,7 @@ class BatchRenderWindow:
                                    lambda w, e: self.launch_render(), 
                                    None)
                                          
-        self.stop_render_button = Gtk.Button(_("Stop Render"))
+        self.stop_render_button = Gtk.Button(label=_("Stop Render"))
         self.stop_render_button.set_sensitive(False)
         self.stop_render_button.connect("clicked", 
                                    lambda w, e: self.abort_render(), 
@@ -854,6 +848,7 @@ class BatchRenderWindow:
         self.stop_render_button.set_sensitive(False)
         self.render_progress_bar.set_text(self.not_rendering_txt)
         self.current_render.set_text("")
+        self.current_file.set_text("")
         self.remove_selected.set_sensitive(True)
         self.remove_finished.set_sensitive(True)
 
@@ -875,7 +870,7 @@ class RenderQueueView(Gtk.VBox):
         self.scroll.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
 
         # View
-        self.treeview = Gtk.TreeView(self.storemodel)
+        self.treeview = Gtk.TreeView(model=self.storemodel)
         self.treeview.set_property("rules_hint", True)
         self.treeview.set_headers_visible(True)
         tree_sel = self.treeview.get_selection()
@@ -1165,6 +1160,7 @@ def add_single_render_item(flowblade_project, render_path, args_vals_list, mark_
                                       sequence_index, args_vals_list, timestamp, length, \
                                       mark_in, mark_out, render_data)
 
+    print("Single render with argsvals:", render_item.args_vals_list)
 
     # Write project 
     project_path = hidden_dir + CURRENT_RENDER_PROJECT_FILE
@@ -1211,49 +1207,32 @@ def single_render_main(root_path):
     # Load editor prefs and list of recent projects
     editorpersistance.load()
     
-    # Init translations module with translations data
-    translations.init_languages()
-    translations.load_filters_translations()
-    mlttransitions.init_module()
+    # Create app.
+    global _single_render_app
+    _single_render_app = SingleRenderApp()
+    _single_render_app.run(None)
 
-    # Init gtk threads
-    Gdk.threads_init()
-    Gdk.threads_enter()
 
-    # Request dark them if so desired
-    editorpersistance.load()
-    if editorpersistance.prefs.theme != appconsts.LIGHT_THEME:
-        Gtk.Settings.get_default().set_property("gtk-application-prefer-dark-theme", True)
-        if editorpersistance.prefs.theme == appconsts.FLOWBLADE_THEME:
-            gui.apply_gtk_css()
+class SingleRenderApp(Gtk.Application):
+    def __init__(self, *args, **kwargs):
+        Gtk.Application.__init__(self, application_id=None,
+                                 flags=Gio.ApplicationFlags.FLAGS_NONE)
+        self.connect("activate", self.on_activate)
 
-    repo = mlt.Factory().init()
-    processutils.prepare_mlt_repo(repo)
+    def on_activate(self, data=None):
+        gui.apply_theme(editorpersistance.prefs.theme)
 
-    # Set numeric locale to use "." as radix, MLT initilizes this to OS locale and this causes bugs 
-    locale.setlocale(locale.LC_NUMERIC, 'C')
+        mltinit.init_with_translations()
+    
+        global single_render_window
+        single_render_window = SingleRenderWindow()
 
-    # Check for codecs and formats on the system
-    mltenv.check_available_features(repo)
-    renderconsumer.load_render_profiles()
-
-    # Load filter and compositor descriptions from xml files.
-    mltfilters.load_filters_xml(mltenv.services)
-    mlttransitions.load_compositors_xml(mltenv.transitions)
-
-    # Create list of available mlt profiles
-    mltprofiles.load_profile_list()
-
-    global single_render_window
-    single_render_window = SingleRenderWindow()
-
-    global single_render_thread
-    single_render_thread = SingleRenderThread()
-    single_render_thread.start()
-
-    Gtk.main()
-    Gdk.threads_leave()
+        global single_render_thread
+        single_render_thread = SingleRenderThread()
+        single_render_thread.start()
 
+        self.add_window(single_render_window.window)
+        
 
 class SingleRenderThread(threading.Thread):
     def __init__(self):
@@ -1281,6 +1260,8 @@ class SingleRenderThread(threading.Thread):
 
         project = persistance.load_project(project_file_path, False)
 
+        project.c_seq.fix_v1_for_render()
+
         producer = project.c_seq.tractor
         profile = mltprofiles.get_profile(render_item.render_data.profile_name)
         
@@ -1312,10 +1293,7 @@ class SingleRenderThread(threading.Thread):
         # Set render start time and item state
         render_item.render_started()
 
-        Gdk.threads_enter()
-        #single_render_window.update_queue_view()
-        single_render_window.current_render.set_text("  " + os.path.basename(render_item.render_path))
-        Gdk.threads_leave()
+        GLib.idle_add(self._show_current_render, render_item)
 
         # Make sure that render thread is actually running before
         # testing render_thread.running value later
@@ -1330,25 +1308,31 @@ class SingleRenderThread(threading.Thread):
             now = time.time()
             current_render_time = now - render_item.start_time
             
-            Gdk.threads_enter()
-            single_render_window.update_render_progress(render_fraction, render_item.get_display_name(), current_render_time)
-            Gdk.threads_leave()
+            GLib.idle_add(self._update_render_progress, render_fraction, render_item.get_display_name(), current_render_time)
             
             if render_thread.running == False: # Rendering has reached end
                 self.running = False
-                
-                Gdk.threads_enter()
-                single_render_window.render_progress_bar.set_fraction(1.0)
-                Gdk.threads_leave()
+
+                GLib.idle_add(self._update_progress_bar, 1.0)
 
             time.sleep(0.33)
                 
         render_thread.shutdown()
         global single_render_thread
         single_render_thread = None
+
         # Update view for render end
         GLib.idle_add(_single_render_shutdown)
 
+    def _show_current_render(self, render_item):
+        single_render_window.current_render.set_text("  " + os.path.basename(render_item.render_path))
+
+    def _update_render_progress(self, fraction, display_name, current_render_time):
+        single_render_window.update_render_progress(fraction, display_name, current_render_time)
+
+    def _update_progress_bar(self, fraction):
+        single_render_window.render_progress_bar.set_fraction(fraction)
+                
     def is_frame_sequence_render(self, vcodec):
         if vcodec in ["png","bmp","dpx","ppm","targa","tiff"]:
             return True
@@ -1400,7 +1384,7 @@ class SingleRenderWindow:
         info_vbox.pack_start(guiutils.get_left_justified_box([current_r_t, self.current_render_time]), False, False, 0)
         info_vbox.pack_start(guiutils.get_left_justified_box([est_r, self.est_time_left]), False, False, 0)
 
-        self.stop_render_button = Gtk.Button(_("Stop Render"))
+        self.stop_render_button = Gtk.Button(label=_("Stop Render"))
         self.stop_render_button.connect("clicked", 
                                    lambda w, e: _start_single_render_shutdown(), 
                                    None)
@@ -1453,5 +1437,4 @@ def _start_single_render_shutdown():
     single_render_thread.abort()
 
 def _single_render_shutdown():
-    Gtk.main_quit()
-    
+    _single_render_app.quit()
